use crate::time::Time;
use std::{
	borrow::Cow,
	collections::VecDeque,
	fmt::{self, Write as _}
};

pub(crate) enum Filter {
	/// Apply an alpha channel on the video. No audio.
	Alpha {
		input: Cow<'static, str>,
		alpha: f32,
		output: Cow<'static, str>
	},

	/// Overlay the one video over the other. The audio is copied.
	Overlay {
		video_input: Cow<'static, str>,
		overlay_input: Cow<'static, str>,
		x: Cow<'static, str>,
		y: Cow<'static, str>,
		repeatlast: bool,
		output: Cow<'static, str>
	},

	/// Concatenate audio and video.
	Concat {
		inputs: VecDeque<Cow<'static, str>>,
		output: Cow<'static, str>
	},

	/// Fade audio and video.
	Fade {
		input: Cow<'static, str>,
		direction: &'static str,
		start: Time,
		duration: Time,
		output: Cow<'static, str>
	},

	/// Fade only video using the alpha channel.
	FadeAlpha {
		input: Cow<'static, str>,
		direction: &'static str,
		start: Time,
		duration: Time,
		output: Cow<'static, str>
	},

	/// Offset the PTS of the video by the amount of seconds.
	VideoOffset {
		input: Cow<'static, str>,
		seconds: Time,
		output: Cow<'static, str>
	},

	/// Generate silence. The video is copied.
	GenerateSilence {
		video: Cow<'static, str>,
		output: Cow<'static, str>
	},

	/// Fast forward.
	FastForward {
		input: Cow<'static, str>,
		ffinput: Cow<'static, str>,
		multiplier: usize,
		output: Cow<'static, str>
	}
}

impl Filter {
	pub(crate) fn append_to_complex_filter(
		&self,
		complex: &mut String,
		filter_idx: &mut usize
	) -> fmt::Result {
		match self {
			Self::Alpha {
				input,
				alpha,
				output
			} => {
				writeln!(
					complex,
					"{}format=yuva444p,colorchannelmixer=aa={alpha}{};",
					channel('v', input),
					channel('v', output)
				)?;
			},

			Self::Overlay {
				video_input,
				overlay_input,
				x,
				y,
				repeatlast,
				output
			} => {
				let repeatlast: u8 = (*repeatlast).into();
				writeln!(
					complex,
					"{}{}overlay=x={x}:y={y}:repeatlast={repeatlast}:eval=init{};",
					channel('v', video_input),
					channel('v', overlay_input),
					channel('v', output)
				)?;
				writeln!(
					complex,
					"{}anull{};",
					channel('a', video_input),
					channel('a', output)
				)?;
			},

			Self::Concat { inputs, output } => {
				for i in inputs {
					write!(complex, "{}{}", channel('v', i), channel('a', i))?;
				}
				writeln!(
					complex,
					"concat=n={}:v=1:a=1{}{};",
					inputs.len(),
					channel('v', output),
					channel('a', output)
				)?;
			},

			Self::Fade {
				input,
				direction,
				start,
				duration,
				output
			} => {
				let args = format!("{direction}:st={start}:d={duration}");
				writeln!(
					complex,
					"{}fade={args}{};",
					channel('v', input),
					channel('v', output)
				)?;
				writeln!(
					complex,
					"{}afade=t={args}{};",
					channel('a', input),
					channel('a', output)
				)?;
			},

			Self::FadeAlpha {
				input,
				direction,
				start,
				duration,
				output
			} => {
				writeln!(
					complex,
					"{}fade={direction}:st={start}:d={duration}:alpha=1{};",
					channel('v', input),
					channel('v', output)
				)?;
			},

			Self::VideoOffset {
				input,
				seconds,
				output
			} => {
				writeln!(
					complex,
					"{}setpts=PTS+{seconds}/TB{};",
					channel('v', input),
					channel('v', output)
				)?;
			},

			Self::GenerateSilence { video, output } => {
				writeln!(
					complex,
					"{}null{};",
					channel('v', video),
					channel('v', output)
				)?;
				writeln!(
					complex,
					"aevalsrc=0:s=48000,pan=stereo|c0=c0|c1=c0{};",
					channel('a', output)
				)?;
			},

			Self::FastForward {
				input,
				ffinput,
				multiplier,
				output
			} => {
				let vff = next_tmp(filter_idx);
				writeln!(
					complex,
					"{}setpts=PTS/{multiplier}{vff};",
					channel('v', input)
				)?;
				writeln!(
					complex,
					"{}atempo={multiplier}{};",
					channel('a', input),
					channel('a', output)
				)?;
				writeln!(
					complex,
					"{vff}{}overlay=x=main_w/2-overlay_w/2:y=main_h/2-overlay_h/2{};",
					channel('v', ffinput),
					channel('v', output)
				)?;
			}
		}

		// add a newline after every filter to ease debugging
		writeln!(complex)
	}
}

pub(super) fn channel(channel: char, id: &str) -> String {
	if id.chars().any(|ch| !ch.is_ascii_digit()) {
		format!("[{channel}_{id}]")
	} else {
		format!("[{id}:{channel}]")
	}
}

fn next_tmp(filter_idx: &mut usize) -> String {
	*filter_idx += 1;
	format!("[tmp{filter_idx}]")
}
